@@ -513,9 +513,7 @@ VOID ReadConfig(CHAR16 *FileName)
           if (SelfVolume->VolName) {
              TempStr = SelfVolume->VolName ? StrDuplicate(SelfVolume->VolName) : NULL;
           } else {
-             TempStr = AllocateZeroPool(256 * sizeof(CHAR16));
-             if (TempStr != NULL)
-                SPrint(TempStr, 255, L"fs%d", SelfVolume->VolNumber);
+             TempStr = GuidAsString(&(SelfVolume->PartGuid));
           } // if/else
        }
        MergeStrings(&TempStr, SelfDirPath, L':');
@@ -523,11 +521,14 @@ VOID ReadConfig(CHAR16 *FileName)
        GlobalConfig.DontScanDirs = TempStr;
        MyFreePool(GlobalConfig.DontScanFiles);
        GlobalConfig.DontScanFiles = StrDuplicate(DONT_SCAN_FILES);
+       MyFreePool(GlobalConfig.DontScanTools);
+       GlobalConfig.DontScanTools = NULL;
        MergeStrings(&(GlobalConfig.DontScanFiles), MOK_NAMES, L',');
        MergeStrings(&(GlobalConfig.DontScanFiles), FWUPDATE_NAMES, L',');
        MyFreePool(GlobalConfig.DontScanVolumes);
        GlobalConfig.DontScanVolumes = StrDuplicate(DONT_SCAN_VOLUMES);
        GlobalConfig.WindowsRecoveryFiles = StrDuplicate(WINDOWS_RECOVERY_FILES);
+       GlobalConfig.MacOSRecoveryFiles = StrDuplicate(MACOS_RECOVERY_FILES);
        if (GlobalConfig.DefaultSelection != NULL) {
           MyFreePool(GlobalConfig.DefaultSelection);
           GlobalConfig.DefaultSelection = NULL;
@@ -620,6 +621,9 @@ VOID ReadConfig(CHAR16 *FileName)
         } else if (MyStriCmp(TokenList[0], L"don't_scan_files") || MyStriCmp(TokenList[0], L"dont_scan_files")) {
            HandleStrings(TokenList, TokenCount, &(GlobalConfig.DontScanFiles));
 
+        } else if (MyStriCmp(TokenList[0], L"don't_scan_tools") || MyStriCmp(TokenList[0], L"dont_scan_tools")) {
+           HandleStrings(TokenList, TokenCount, &(GlobalConfig.DontScanTools));
+
         } else if (MyStriCmp(TokenList[0], L"windows_recovery_files")) {
            HandleStrings(TokenList, TokenCount, &(GlobalConfig.WindowsRecoveryFiles));
 
@@ -628,6 +632,7 @@ VOID ReadConfig(CHAR16 *FileName)
 
         } else if (MyStriCmp(TokenList[0], L"showtools")) {
             SetMem(GlobalConfig.ShowTools, NUM_TOOLS * sizeof(UINTN), 0);
+            GlobalConfig.HiddenTags = FALSE;
             for (i = 1; (i < TokenCount) && (i < NUM_TOOLS); i++) {
                 FlagName = TokenList[i];
                 if (MyStriCmp(FlagName, L"shell")) {
@@ -660,6 +665,9 @@ VOID ReadConfig(CHAR16 *FileName)
                    GlobalConfig.ShowTools[i - 1] = TAG_MEMTEST;
                 } else if (MyStriCmp(FlagName, L"netboot")) {
                    GlobalConfig.ShowTools[i - 1] = TAG_NETBOOT;
+                } else if (MyStriCmp(FlagName, L"hidden_tags")) {
+                    GlobalConfig.ShowTools[i - 1] = TAG_HIDDEN;
+                    GlobalConfig.HiddenTags = TRUE;
                 } else {
                    Print(L" unknown showtools flag: '%s'\n", FlagName);
                 }
@@ -679,14 +687,23 @@ VOID ReadConfig(CHAR16 *FileName)
 
         } else if (MyStriCmp(TokenList[0], L"small_icon_size") && (TokenCount == 2)) {
            HandleInt(TokenList, TokenCount, &i);
-           if (i >= 32)
+           if (i >= 32) {
               GlobalConfig.IconSizes[ICON_SIZE_SMALL] = i;
+              HaveResized = TRUE;
+           }
 
         } else if (MyStriCmp(TokenList[0], L"big_icon_size") && (TokenCount == 2)) {
            HandleInt(TokenList, TokenCount, &i);
            if (i >= 32) {
               GlobalConfig.IconSizes[ICON_SIZE_BIG] = i;
               GlobalConfig.IconSizes[ICON_SIZE_BADGE] = i / 4;
+              HaveResized = TRUE;
+           }
+
+        } else if (MyStriCmp(TokenList[0], L"mouse_size") && (TokenCount == 2)) {
+           HandleInt(TokenList, TokenCount, &i);
+           if (i >= DEFAULT_MOUSE_SIZE) {
+              GlobalConfig.IconSizes[ICON_SIZE_MOUSE] = i;
            }
 
         } else if (MyStriCmp(TokenList[0], L"selection_small")) {
@@ -744,6 +761,9 @@ VOID ReadConfig(CHAR16 *FileName)
         } else if (MyStriCmp(TokenList[0], L"fold_linux_kernels")) {
             GlobalConfig.FoldLinuxKernels = HandleBoolean(TokenList, TokenCount);
 
+        } else if (MyStriCmp(TokenList[0], L"extra_kernel_version_strings")) {
+            HandleStrings(TokenList, TokenCount, &(GlobalConfig.ExtraKernelVersionStrings));
+
         } else if (MyStriCmp(TokenList[0], L"max_tags")) {
            HandleInt(TokenList, TokenCount, &(GlobalConfig.MaxTags));
 
@@ -761,8 +781,25 @@ VOID ReadConfig(CHAR16 *FileName)
               ReadConfig(TokenList[1]);
            }
 
+        } else if (MyStriCmp(TokenList[0], L"enable_mouse")) {
+           GlobalConfig.EnableMouse = HandleBoolean(TokenList, TokenCount);
+           if(GlobalConfig.EnableMouse) {
+               GlobalConfig.EnableTouch = FALSE;
+           }
+        
         } else if (MyStriCmp(TokenList[0], L"enable_touch")) {
            GlobalConfig.EnableTouch = HandleBoolean(TokenList, TokenCount);
+           if(GlobalConfig.EnableTouch) {
+               GlobalConfig.EnableMouse = FALSE;
+           }
+           
+        } else if (MyStriCmp(TokenList[0], L"mouse_speed") && (TokenCount == 2)) {
+           HandleInt(TokenList, TokenCount, &i);
+           if (i < 1)
+              i = 1;
+           if (i > 32)
+              i = 32;
+           GlobalConfig.MouseSpeed = i;
         }
 
         FreeTokenLine(&TokenList, &TokenCount);
@@ -777,50 +814,6 @@ VOID ReadConfig(CHAR16 *FileName)
     }
 } /* VOID ReadConfig() */
 
-// Finds a volume with the specified Identifier (a filesystem label, a
-// partition name, a partition GUID, or a number followed by a colon). If
-// found, sets *Volume to point to that volume. If not, leaves it unchanged.
-// Returns TRUE if a match was found, FALSE if not.
-static BOOLEAN FindVolume(REFIT_VOLUME **Volume, CHAR16 *Identifier) {
-   UINTN     i = 0, CountedVolumes = 0, Length;
-   INTN      Number = -1;
-   BOOLEAN   Found = FALSE, IdIsGuid = FALSE;
-   EFI_GUID  VolGuid, NullGuid = NULL_GUID_VALUE;
-
-   VolGuid = StringAsGuid(Identifier);
-   Length = StrLen(Identifier);
-   if ((Length >= 2) && (Identifier[Length - 1] == L':') &&
-       (Identifier[0] >= L'0') && (Identifier[0] <= L'9')) {
-      Number = (INTN) Atoi(Identifier);
-   } else if (IsGuid(Identifier)) {
-      IdIsGuid = TRUE;
-   }
-   while ((i < VolumesCount) && (!Found)) {
-      if (Number >= 0) { // User specified a volume by number
-         if (Volumes[i]->IsReadable) {
-            if (CountedVolumes == Number) {
-               *Volume = Volumes[i];
-               Found = TRUE;
-            }
-            CountedVolumes++;
-         } // if
-      } else { // User specified a volume by label or GUID
-         if (MyStriCmp(Identifier, Volumes[i]->VolName) || MyStriCmp(Identifier, Volumes[i]->PartName)) {
-            *Volume = Volumes[i];
-            Found = TRUE;
-         } // if
-         if (IdIsGuid && !Found) {
-            if (GuidsAreEqual(&VolGuid, &(Volumes[i]->PartGuid)) && !GuidsAreEqual(&NullGuid, &(Volumes[i]->PartGuid))) {
-               *Volume = Volumes[i];
-               Found = TRUE;
-            } // if
-         } // if
-      } // if/else
-      i++;
-   } // while()
-   return (Found);
-} // static VOID FindVolume()
-
 static VOID AddSubmenu(LOADER_ENTRY *Entry, REFIT_FILE *File, REFIT_VOLUME *Volume, CHAR16 *Title) {
    REFIT_MENU_SCREEN  *SubScreen;
    LOADER_ENTRY       *SubEntry;
@@ -841,7 +834,7 @@ static VOID AddSubmenu(LOADER_ENTRY *Entry, REFIT_FILE *File, REFIT_VOLUME *Volu
       if (MyStriCmp(TokenList[0], L"loader") && (TokenCount > 1)) { // set the boot loader filename
          MyFreePool(SubEntry->LoaderPath);
          SubEntry->LoaderPath = StrDuplicate(TokenList[1]);
-         SubEntry->DevicePath = FileDevicePath(Volume->DeviceHandle, SubEntry->LoaderPath);
+         SubEntry->Volume = Volume;
 
       } else if (MyStriCmp(TokenList[0], L"volume") && (TokenCount > 1)) {
          if (FindVolume(&Volume, TokenList[1])) {
@@ -850,7 +843,7 @@ static VOID AddSubmenu(LOADER_ENTRY *Entry, REFIT_FILE *File, REFIT_VOLUME *Volu
                SubEntry->me.Title        = AllocateZeroPool(256 * sizeof(CHAR16));
                SPrint(SubEntry->me.Title, 255, L"Boot %s from %s", (Title != NULL) ? Title : L"Unknown", Volume->VolName);
                SubEntry->me.BadgeImage   = Volume->VolBadgeImage;
-               SubEntry->VolName         = Volume->VolName;
+               SubEntry->Volume          = Volume;
             } // if volume is readable
          } // if match found
 
@@ -913,14 +906,14 @@ static LOADER_ENTRY * AddStanzaEntries(REFIT_FILE *File, REFIT_VOLUME *Volume, C
    SPrint(Entry->me.Title, 255, L"Boot %s from %s", (Title != NULL) ? Title : L"Unknown", CurrentVolume->VolName);
    Entry->me.Row          = 0;
    Entry->me.BadgeImage   = CurrentVolume->VolBadgeImage;
-   Entry->VolName         = CurrentVolume->VolName;
+   Entry->Volume          = CurrentVolume;
+   Entry->DiscoveryType   = DISCOVERY_TYPE_MANUAL;
 
    // Parse the config file to add options for a single stanza, terminating when the token
    // is "}" or when the end of file is reached.
    while (((TokenCount = ReadTokenLine(File, &TokenList)) > 0) && (StrCmp(TokenList[0], L"}") != 0)) {
       if (MyStriCmp(TokenList[0], L"loader") && (TokenCount > 1)) { // set the boot loader filename
          Entry->LoaderPath = StrDuplicate(TokenList[1]);
-         Entry->DevicePath = FileDevicePath(CurrentVolume->DeviceHandle, Entry->LoaderPath);
          SetLoaderDefaults(Entry, TokenList[1], CurrentVolume);
          MyFreePool(Entry->LoadOptions);
          Entry->LoadOptions = NULL; // Discard default options, if any
@@ -933,7 +926,7 @@ static LOADER_ENTRY * AddStanzaEntries(REFIT_FILE *File, REFIT_VOLUME *Volume, C
                Entry->me.Title        = AllocateZeroPool(256 * sizeof(CHAR16));
                SPrint(Entry->me.Title, 255, L"Boot %s from %s", (Title != NULL) ? Title : L"Unknown", CurrentVolume->VolName);
                Entry->me.BadgeImage   = CurrentVolume->VolBadgeImage;
-               Entry->VolName         = CurrentVolume->VolName;
+               Entry->Volume          = CurrentVolume;
             } // if volume is readable
          } // if match found
 
@@ -1139,14 +1132,9 @@ static REFIT_FILE * GenerateOptionsFromPartTypes(VOID) {
 // you pass this function the filename of the Linux kernel, initial RAM disk, or other
 // file in the target directory, and this function finds the file with a name in the
 // comma-delimited list of names specified by LINUX_OPTIONS_FILENAMES within that
-// directory and loads it. This function tries multiple files because I originally
-// used the filename linux.conf, but close on the heels of that decision, the Linux
-// kernel developers decided to use that name for a similar purpose, but with a
-// different file format. Thus, I'm migrating rEFInd to use the name refind_linux.conf,
-// but I want a migration period in which both names are used.
-// If a rEFInd options file can't be found, try to generate minimal options from
-// /etc/fstab on the same volume as the kernel. This typically works only if the
-// kernel is being read from the Linux root filesystem.
+// directory and loads it. If a rEFInd options file can't be found, try to generate
+// minimal options from /etc/fstab on the same volume as the kernel. This typically
+// works only if the kernel is being read from the Linux root filesystem.
 //
 // The return value is a pointer to the REFIT_FILE handle for the file, or NULL if
 // it wasn't found.
