@@ -53,7 +53,7 @@
 #endif
 #include "../EfiLib/GenericBdsLib.h"
 
-#include "libeg.h"
+#include "../libeg/libeg.h"
 
 #define REFIT_DEBUG (0)
 
@@ -77,7 +77,8 @@
 #define TAG_NETBOOT          (17)
 #define TAG_CSR_ROTATE       (18)
 #define TAG_FWUPDATE_TOOL    (19)
-#define NUM_TOOLS            (20)
+#define TAG_HIDDEN           (20)
+#define NUM_TOOLS            (21)
 
 #define NUM_SCAN_OPTIONS 10
 
@@ -90,15 +91,16 @@
 #define GRAPHICS_FOR_GRUB       8
 #define GRAPHICS_FOR_WINDOWS   16
 
-// Load types
-#define TYPE_EFI    1
-#define TYPE_LEGACY 2
-
 // Type of legacy (BIOS) boot support detected
 #define LEGACY_TYPE_NONE 0
 #define LEGACY_TYPE_MAC  1
 #define LEGACY_TYPE_UEFI 2
 
+// How was a loader added to the menu?
+#define DISCOVERY_TYPE_UNKNOWN  0
+#define DISCOVERY_TYPE_AUTO     1
+#define DISCOVERY_TYPE_MANUAL   2
+
 #ifdef __MAKEWITH_GNUEFI
 //
 // define BBS Device Types
@@ -143,14 +145,27 @@
 // Sizes of the default icons; badges are 1/4 the big icon size
 #define DEFAULT_SMALL_ICON_SIZE 48
 #define DEFAULT_BIG_ICON_SIZE   128
+#define DEFAULT_MOUSE_SIZE      16
 
 // Codes for types of icon sizes; used for indexing into GlobalConfig.IconSizes[]
 #define ICON_SIZE_BADGE 0
 #define ICON_SIZE_SMALL 1
 #define ICON_SIZE_BIG   2
+#define ICON_SIZE_MOUSE 3
+
+// Minimum horizontal resolution for a screen to be consider high-DPI
+#define HIDPI_MIN 1921
 
 // Names of binaries that can manage MOKs....
+#if defined (EFIX64)
+#define MOK_NAMES               L"MokManager.efi,HashTool.efi,HashTool-signed.efi,KeyTool.efi,KeyTool-signed.efi,mmx64.efi"
+#elif defined(EFI32)
+#define MOK_NAMES               L"MokManager.efi,HashTool.efi,HashTool-signed.efi,KeyTool.efi,KeyTool-signed.efi,mmia32.efi"
+#elif defined(EFIAARCH64)
+#define MOK_NAMES               L"MokManager.efi,HashTool.efi,HashTool-signed.efi,KeyTool.efi,KeyTool-signed.efi,mmaa64.efi"
+#else
 #define MOK_NAMES               L"MokManager.efi,HashTool.efi,HashTool-signed.efi,KeyTool.efi,KeyTool-signed.efi"
+#endif
 // Names of binaries that can update firmware....
 #if defined (EFIX64)
 #define FWUPDATE_NAMES          L"fwupx64.efi"
@@ -167,7 +182,9 @@
 // Directories to search for memtest86....
 #define MEMTEST_LOCATIONS       L"EFI\\tools,EFI\\tools\\memtest86,EFI\\tools\\memtest,EFI\\memtest86,EFI\\memtest"
 // Files that may be Windows recovery files
-#define WINDOWS_RECOVERY_FILES  L"EFI\\Microsoft\\Boot\\LrsBootmgr.efi,Recovery:\\EFI\\BOOT\\bootx64.efi,Recovery:\\EFI\\BOOT\\bootia32.efi"
+#define WINDOWS_RECOVERY_FILES  L"EFI\\Microsoft\\Boot\\LrsBootmgr.efi,Recovery:\\EFI\\BOOT\\bootx64.efi,Recovery:\\EFI\\BOOT\\bootia32.efi,\\EFI\\OEM\\Boot\\bootmgfw.efi"
+// Files that may be macOS recovery files
+#define MACOS_RECOVERY_FILES    L"com.apple.recovery.boot\\boot.efi"
 
 // Filename patterns that identify EFI boot loaders. Note that a single case (either L"*.efi" or
 // L"*.EFI") is fine for most systems; but Gigabyte's buggy Hybrid EFI does a case-sensitive
@@ -197,6 +214,9 @@
 #define NULL_GUID_VALUE { 0x00000000, 0x0000, 0x0000, {0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00} };
 #define REFIND_GUID_VALUE { 0x36D08FA7, 0xCF0B, 0x42F5, {0x8F, 0x14, 0x68, 0xDF, 0x73, 0xED, 0x37, 0x40} };
 
+// Configuration file variables
+#define KERNEL_VERSION L"%v"
+
 //
 // global definitions
 //
@@ -231,11 +251,10 @@ typedef struct {
    EFI_GUID            PartGuid;
    EFI_GUID            PartTypeGuid;
    BOOLEAN             IsMarkedReadOnly;
-   UINTN               VolNumber;
    EG_IMAGE            *VolIconImage;
    EG_IMAGE            *VolBadgeImage;
    UINTN               DiskKind;
-   BOOLEAN             IsAppleLegacy;
+//    BOOLEAN             IsAppleLegacy;
    BOOLEAN             HasBootCode;
    CHAR16              *OSIconName;
    CHAR16              *OSName;
@@ -278,13 +297,13 @@ typedef struct {
    REFIT_MENU_ENTRY me;
    CHAR16           *Title;
    CHAR16           *LoaderPath;
-   CHAR16           *VolName;
-   EFI_DEVICE_PATH  *DevicePath;
+   REFIT_VOLUME     *Volume;
    BOOLEAN          UseGraphicsMode;
    BOOLEAN          Enabled;
    CHAR16           *LoadOptions;
    CHAR16           *InitrdPath; // Linux stub loader only
    CHAR8            OSType;
+   UINTN            DiscoveryType;
 } LOADER_ENTRY;
 
 typedef struct {
@@ -301,7 +320,9 @@ typedef struct {
    BOOLEAN          DeepLegacyScan;
    BOOLEAN          EnableAndLockVMX;
    BOOLEAN          FoldLinuxKernels;
+   BOOLEAN          EnableMouse;
    BOOLEAN          EnableTouch;
+   BOOLEAN          HiddenTags;
    UINTN            RequestedScreenWidth;
    UINTN            RequestedScreenHeight;
    UINTN            BannerBottomEdge;
@@ -313,7 +334,8 @@ typedef struct {
    UINTN            LegacyType;
    UINTN            ScanDelay;
    UINTN            ScreensaverTime;
-   UINTN            IconSizes[3];
+   UINTN            MouseSpeed;
+   UINTN            IconSizes[4];
    UINTN            BannerScale;
    REFIT_VOLUME     *DiscoveredRoot;
    EFI_DEVICE_PATH  *SelfDevicePath;
@@ -327,9 +349,12 @@ typedef struct {
    CHAR16           *DontScanVolumes;
    CHAR16           *DontScanDirs;
    CHAR16           *DontScanFiles;
+   CHAR16           *DontScanTools;
    CHAR16           *WindowsRecoveryFiles;
+   CHAR16           *MacOSRecoveryFiles;
    CHAR16           *DriverDirs;
    CHAR16           *IconsDir;
+   CHAR16           *ExtraKernelVersionStrings;
    CHAR16           *SpoofOSXVersion;
    UINT32_LIST      *CsrValues;
    UINTN            ShowTools[NUM_TOOLS];
@@ -353,16 +378,13 @@ extern REFIT_CONFIG     GlobalConfig;
 extern EFI_GUID gEfiLegacyBootProtocolGuid;
 extern EFI_GUID gEfiGlobalVariableGuid;
 
-EFI_STATUS StartEFIImageList(IN EFI_DEVICE_PATH **DevicePaths,
-                             IN CHAR16 *LoadOptions, IN UINTN LoaderType,
-                             IN CHAR16 *ImageTitle, IN CHAR8 OSType,
-                             OUT UINTN *ErrorInStep,
-                             IN BOOLEAN Verbose,
-                             IN BOOLEAN IsDriver);
-EFI_STATUS StartEFIImage(IN EFI_DEVICE_PATH *DevicePath,
-                         IN CHAR16 *LoadOptions, IN UINTN LoaderType,
-                         IN CHAR16 *ImageTitle, IN CHAR8 OSType,
-                         OUT UINTN *ErrorInStep,
+extern BOOLEAN HaveResized;
+
+EFI_STATUS StartEFIImage(IN REFIT_VOLUME *Volume,
+                         IN CHAR16 *Filename,
+                         IN CHAR16 *LoadOptions,
+                         IN CHAR16 *ImageTitle,
+                         IN CHAR8 OSType,
                          IN BOOLEAN Verbose,
                          IN BOOLEAN IsDriver);
 LOADER_ENTRY *InitializeLoaderEntry(IN LOADER_ENTRY *Entry);
@@ -373,6 +395,7 @@ LOADER_ENTRY * MakeGenericLoaderEntry(VOID);
 VOID SetLoaderDefaults(LOADER_ENTRY *Entry, CHAR16 *LoaderPath, IN REFIT_VOLUME *Volume);
 LOADER_ENTRY * AddPreparedLoaderEntry(LOADER_ENTRY *Entry);
 VOID StoreLoaderName(IN CHAR16 *Name);
+VOID RescanAll(BOOLEAN DisplayMessage);
 
 #endif
 
