@@ -34,7 +34,7 @@
  * OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
  */
 /*
- * Modifications copyright (c) 2012-2015 Roderick W. Smith
+ * Modifications copyright (c) 2012-2017 Roderick W. Smith
  *
  * Modifications distributed under the terms of the GNU General Public
  * License (GPL) version 3 (GPLv3), or (at your option) any later version.
@@ -120,37 +120,38 @@ extern GPT_DATA *gPartitions;
 // in pathnames, because some user inputs can produce duplicate directory
 // separators, and because we want consistent start and end slashes for
 // directory comparisons. A special case: If the PathName refers to root,
-// return "/", since some firmware implementations flake out if this
-// isn't present.
+// but is non-empty, return "\", since some firmware implementations flake
+// out if this isn't present.
 VOID CleanUpPathNameSlashes(IN OUT CHAR16 *PathName) {
-    CHAR16   *NewName;
-    UINTN    i, Length, FinalChar = 0;
-    BOOLEAN  LastWasSlash = FALSE;
+    UINTN Source = 0, Dest = 0;
 
-    Length = StrLen(PathName);
-    NewName = AllocateZeroPool(sizeof(CHAR16) * (Length + 2));
-    if (NewName != NULL) {
-        for (i = 0; i < Length; i++) {
-            if ((PathName[i] == L'/') || (PathName[i] == L'\\')) {
-                if ((!LastWasSlash) && (FinalChar != 0))
-                    NewName[FinalChar++] = L'\\';
-                LastWasSlash = TRUE;
+    if ((PathName == NULL) || (PathName[0] == '\0'))
+        return;
+
+    while (PathName[Source] != '\0') {
+        if ((PathName[Source] == L'/') || (PathName[Source] == L'\\')) {
+            if (Dest == 0) { // Skip slash if to first position
+                Source++;
             } else {
-                NewName[FinalChar++] = PathName[i];
-                LastWasSlash = FALSE;
+                PathName[Dest] = L'\\';
+                do { // Skip subsequent slashes
+                    Source++;
+                } while ((PathName[Source] == L'/') || (PathName[Source] == L'\\'));
+                Dest++;
             } // if/else
-        } // for
-        NewName[FinalChar] = 0;
-        if ((FinalChar > 0) && (NewName[FinalChar - 1] == L'\\'))
-            NewName[--FinalChar] = 0;
-        if (FinalChar == 0) {
-            NewName[0] = L'\\';
-            NewName[1] = 0;
-        }
-        // Copy the transformed name back....
-        StrCpy(PathName, NewName);
-        FreePool(NewName);
-    } // if allocation OK
+        } else { // Regular character; copy it straight....
+            PathName[Dest] = PathName[Source];
+            Source++;
+            Dest++;
+        } // if/else
+    } // while()
+    if ((Dest > 0) && (PathName[Dest - 1] == L'\\'))
+        Dest--;
+    PathName[Dest] = L'\0';
+    if (PathName[0] == L'\0') {
+        PathName[0] = L'\\';
+        PathName[1] = L'\0';
+    }
 } // CleanUpPathNameSlashes()
 
 // Splits an EFI device path into device and filename components. For instance, if InString is
@@ -211,7 +212,7 @@ static EFI_STATUS FinishInitRefitLib(VOID)
 EFI_STATUS InitRefitLib(IN EFI_HANDLE ImageHandle)
 {
     EFI_STATUS  Status;
-    CHAR16      *DevicePathAsString, *Temp;
+    CHAR16      *DevicePathAsString, *Temp = NULL;
 
     SelfImageHandle = ImageHandle;
     Status = refit_call3_wrapper(BS->HandleProtocol, SelfImageHandle, &LoadedImageProtocol, (VOID **) &SelfLoadedImage);
@@ -495,7 +496,7 @@ static VOID SetFilesystemData(IN UINT8 *Buffer, IN UINTN BufferSize, IN OUT REFI
             } else { // none of these features; presume it's ext2...
                Volume->FSType = FS_TYPE_EXT2;
             }
-            CopyMem(&(Volume->VolUuid), Buffer + 1024 + 104, sizeof(EFI_GUID));
+            CopyMem(&(Volume->VolUuid), Buffer + 1024 + 120, sizeof(EFI_GUID));
             return;
          }
       } // search for ext2/3/4 magic
@@ -865,7 +866,12 @@ static VOID SetPartGuidAndName(REFIT_VOLUME *Volume, EFI_DEVICE_PATH_PROTOCOL *D
                 } // if (GUIDs match && automounting OK)
                 Volume->IsMarkedReadOnly = ((PartInfo->attributes & GPT_READ_ONLY) > 0);
             } // if (PartInfo exists)
-        } // if (GPT disk)
+        } else {
+            // TODO: Better to assign a random GUID to MBR partitions, but I couldn't
+            // find an EFI function to do this. The below GUID is just one that I
+            // generated in Linux.
+            Volume->PartGuid = StringAsGuid(L"92a6c61f-7130-49b9-b05c-8d7e7b039127");
+        } // if/else (GPT disk)
     } // if (disk device)
 } // VOID SetPartGuid()
 
@@ -940,11 +946,11 @@ VOID ScanVolume(REFIT_VOLUME *Volume)
             Bootable = TRUE;
         }
 
-        if (DevicePathType(DevicePath) == MEDIA_DEVICE_PATH && DevicePathSubType(DevicePath) == MEDIA_VENDOR_DP) {
-            Volume->IsAppleLegacy = TRUE;             // legacy BIOS device entry
-            // TODO: also check for Boot Camp GUID
-            Bootable = FALSE;   // this handle's BlockIO is just an alias for the whole device
-        }
+//         if (DevicePathType(DevicePath) == MEDIA_DEVICE_PATH && DevicePathSubType(DevicePath) == MEDIA_VENDOR_DP) {
+//             Volume->IsAppleLegacy = TRUE;             // legacy BIOS device entry
+//             // TODO: also check for Boot Camp GUID
+//             Bootable = FALSE;   // this handle's BlockIO is just an alias for the whole device
+//         }
 
         if (DevicePathType(DevicePath) == MESSAGING_DEVICE_PATH) {
             // make a device path for the whole device
@@ -1082,7 +1088,7 @@ VOID ScanVolumes(VOID)
     UINTN                   HandleIndex;
     UINTN                   VolumeIndex, VolumeIndex2;
     UINTN                   PartitionIndex;
-    UINTN                   SectorSum, i, VolNumber = 0;
+    UINTN                   SectorSum, i;
     UINT8                   *SectorBuffer1, *SectorBuffer2;
     EFI_GUID                *UuidList;
     EFI_GUID                NullUuid = NULL_GUID_VALUE;
@@ -1116,10 +1122,6 @@ VOID ScanVolumes(VOID)
               } // if
            } // for
         } // if
-        if (Volume->IsReadable)
-           Volume->VolNumber = VolNumber++;
-        else
-           Volume->VolNumber = VOL_UNREADABLE;
 
         AddListElement((VOID ***) &Volumes, &VolumesCount, Volume);
 
@@ -1407,7 +1409,7 @@ EFI_STATUS DirIterClose(IN OUT REFIT_DIR_ITER *DirIter)
       FreePool(DirIter->LastFileInfo);
       DirIter->LastFileInfo = NULL;
    }
-   if (DirIter->CloseDirHandle)
+   if ((DirIter->CloseDirHandle) && (DirIter->DirHandle->Close))
       refit_call1_wrapper(DirIter->DirHandle->Close, DirIter->DirHandle);
    return DirIter->LastStatus;
 }
@@ -1563,6 +1565,9 @@ VOID FindVolumeAndFilename(IN EFI_DEVICE_PATH *loadpath, OUT REFIT_VOLUME **Devi
     UINTN i = 0;
     BOOLEAN Found = FALSE;
 
+    if (!loadpath || !DeviceVolume || !loader)
+        return;
+
     MyFreePool(*loader);
     MyFreePool(*DeviceVolume);
     *DeviceVolume = NULL;
@@ -1570,6 +1575,10 @@ VOID FindVolumeAndFilename(IN EFI_DEVICE_PATH *loadpath, OUT REFIT_VOLUME **Devi
     *loader = SplitDeviceString(DeviceString);
 
     while ((i < VolumesCount) && (!Found)) {
+        if (Volumes[i]->DevicePath == NULL) {
+            i++;
+            continue;
+        }
         VolumeDeviceString = DevicePathToStr(Volumes[i]->DevicePath);
         Temp = SplitDeviceString(VolumeDeviceString);
         if (MyStriCmp(DeviceString, VolumeDeviceString)) {
@@ -1647,6 +1656,39 @@ VOID SplitPathName(CHAR16 *InPath, CHAR16 **VolName, CHAR16 **Path, CHAR16 **Fil
     MyFreePool(Temp);
 } // VOID SplitPathName()
 
+// Finds a volume with the specified Identifier (a filesystem label, a
+// partition name, or a partition GUID). If found, sets *Volume to point
+// to that volume. If not, leaves it unchanged.
+// Returns TRUE if a match was found, FALSE if not.
+BOOLEAN FindVolume(REFIT_VOLUME **Volume, CHAR16 *Identifier) {
+    UINTN     i = 0;
+    BOOLEAN   Found = FALSE;
+
+    while ((i < VolumesCount) && (!Found)) {
+        if (VolumeMatchesDescription(Volumes[i], Identifier)) {
+            *Volume = Volumes[i];
+            Found = TRUE;
+        } // if
+        i++;
+    } // while()
+    return (Found);
+} // static VOID FindVolume()
+
+// Returns TRUE if Description matches Volume's VolName, PartName, or (once
+// transformed) PartGuid fields, FALSE otherwise (or if either pointer is NULL)
+BOOLEAN VolumeMatchesDescription(REFIT_VOLUME *Volume, CHAR16 *Description) {
+    EFI_GUID TargetVolGuid = NULL_GUID_VALUE;
+
+    if ((Volume == NULL) || (Description == NULL))
+        return FALSE;
+    if (IsGuid(Description)) {
+        TargetVolGuid = StringAsGuid(Description);
+        return GuidsAreEqual(&TargetVolGuid, &(Volume->PartGuid));
+    } else {
+        return (MyStriCmp(Description, Volume->VolName) || MyStriCmp(Description, Volume->PartName));
+    }
+} // BOOLEAN VolumeMatchesDescription()
+
 // Returns TRUE if specified Volume, Directory, and Filename correspond to an
 // element in the comma-delimited List, FALSE otherwise. Note that Directory and
 // Filename must *NOT* include a volume or path specification (that's part of
@@ -1662,44 +1704,21 @@ BOOLEAN FilenameIn(REFIT_VOLUME *Volume, CHAR16 *Directory, CHAR16 *Filename, CH
         while (!Found && (OneElement = FindCommaDelimited(List, i++))) {
             Found = TRUE;
             SplitPathName(OneElement, &TargetVolName, &TargetPath, &TargetFilename);
-            VolumeNumberToName(Volume, &TargetVolName);
-            if (((TargetVolName != NULL) && ((Volume == NULL) || (!MyStriCmp(TargetVolName, Volume->VolName)))) ||
+            if (((TargetVolName != NULL) && (!VolumeMatchesDescription(Volume, TargetVolName))) ||
                  ((TargetPath != NULL) && (!MyStriCmp(TargetPath, Directory))) ||
                  ((TargetFilename != NULL) && (!MyStriCmp(TargetFilename, Filename)))) {
                 Found = FALSE;
             } // if
             MyFreePool(OneElement);
         } // while
+        MyFreePool(TargetVolName);
+        MyFreePool(TargetPath);
+        MyFreePool(TargetFilename);
     } // if
 
-    MyFreePool(TargetVolName);
-    MyFreePool(TargetPath);
-    MyFreePool(TargetFilename);
     return Found;
 } // BOOLEAN FilenameIn()
 
-// If *VolName is of the form "fs#", where "#" is a number, and if Volume points
-// to this volume number, returns with *VolName changed to the volume name, as
-// stored in the Volume data structure.
-// Returns TRUE if this substitution was made, FALSE otherwise.
-BOOLEAN VolumeNumberToName(REFIT_VOLUME *Volume, CHAR16 **VolName) {
-   BOOLEAN MadeSubstitution = FALSE;
-   UINTN VolNum;
-
-   if ((VolName == NULL) || (*VolName == NULL))
-      return FALSE;
-
-   if ((StrLen(*VolName) > 2) && (*VolName[0] == L'f') && (*VolName[1] == L's') && (*VolName[2] >= L'0') && (*VolName[2] <= L'9')) {
-      VolNum = Atoi(*VolName + 2);
-      if (VolNum == Volume->VolNumber) {
-         MyFreePool(*VolName);
-         *VolName = StrDuplicate(Volume->VolName);
-         MadeSubstitution = TRUE;
-      } // if
-   } // if
-   return MadeSubstitution;
-} // BOOLEAN VolumeMatchesNumber()
-
 // Implement FreePool the way it should have been done to begin with, so that
 // it doesn't throw an ASSERT message if fed a NULL pointer....
 VOID MyFreePool(IN VOID *Pointer) {
