@@ -26,6 +26,7 @@
 
 #include "mystrings.h"
 #include "lib.h"
+#include "screen.h"
 
 BOOLEAN StriSubCmp(IN CHAR16 *SmallStr, IN CHAR16 *BigStr) {
     BOOLEAN Found = 0, Terminate = 0;
@@ -236,16 +237,27 @@ BOOLEAN LimitStringLength(CHAR16 *TheString, UINTN Limit) {
 
 // Returns all the digits in the input string, including intervening
 // non-digit characters. For instance, if InString is "foo-3.3.4-7.img",
-// this function returns "3.3.4-7". If InString contains no digits,
-// the return value is NULL.
+// this function returns "3.3.4-7". The GlobalConfig.ExtraKernelVersionStrings
+// variable specifies extra strings that may be treated as numbers. If
+// InString contains no digits or ExtraKernelVersionStrings, the return value
+// is NULL.
 CHAR16 *FindNumbers(IN CHAR16 *InString) {
-    UINTN i, StartOfElement, EndOfElement = 0, CopyLength;
-    CHAR16 *Found = NULL;
+    UINTN i = 0, StartOfElement, EndOfElement = 0, CopyLength;
+    CHAR16 *Found = NULL, *ExtraFound = NULL, *LookFor;
 
     if (InString == NULL)
         return NULL;
 
     StartOfElement = StrLen(InString);
+
+    // Find extra_kernel_version_strings
+    while ((ExtraFound == NULL) && (LookFor = FindCommaDelimited(GlobalConfig.ExtraKernelVersionStrings, i++))) {
+        if ((ExtraFound = MyStrStr(InString, LookFor))) {
+            StartOfElement = ExtraFound - InString;
+            EndOfElement = StartOfElement + StrLen(LookFor) - 1;
+        } // if
+    } // while
+
     // Find start & end of target element
     for (i = 0; InString[i] != L'\0'; i++) {
         if ((InString[i] >= L'0') && (InString[i] <= L'9')) {
@@ -267,6 +279,19 @@ CHAR16 *FindNumbers(IN CHAR16 *InString) {
     return (Found);
 } // CHAR16 *FindNumbers()
 
+// Returns the number of characters that are in common between
+// String1 and String2 before they diverge. For instance, if
+// String1 is "FooBar" and String2 is "FoodiesBar", this function
+// will return "3", since they both start with "Foo".
+UINTN NumCharsInCommon(IN CHAR16* String1, IN CHAR16* String2) {
+    UINTN Count = 0;
+    if ((String1 == NULL) || (String2 == NULL))
+        return 0;
+    while ((String1[Count] != L'\0') && (String2[Count] != L'\0') && (String1[Count] == String2[Count]))
+        Count++;
+    return Count;
+} // UINTN NumCharsInCommon()
+
 // Find the #Index element (numbered from 0) in a comma-delimited string
 // of elements.
 // Returns the found element, or NULL if Index is out of range or InString
@@ -302,6 +327,34 @@ CHAR16 *FindCommaDelimited(IN CHAR16 *InString, IN UINTN Index) {
     return (FoundString);
 } // CHAR16 *FindCommaDelimited()
 
+// Delete an individual element from a comma-separated value list.
+// This function modifies the original *List string, but not the
+// *ToDelete string!
+// Returns TRUE if the item was deleted, FALSE otherwise.
+BOOLEAN DeleteItemFromCsvList(CHAR16 *ToDelete, CHAR16 *List) {
+    CHAR16 *Found, *Comma;
+
+    if ((ToDelete == NULL) || (List == NULL))
+        return FALSE;
+
+    if ((Found = MyStrStr(List, ToDelete)) != NULL) {
+        if ((Comma = MyStrStr(Found, L",")) == NULL) {
+            // Found is final element
+            if (Found == List) { // Found is ONLY element
+                List[0] = L'\0';
+            } else { // Delete the comma preceding Found....
+                Found--;
+                Found[0] = L'\0';
+            } // if/else
+        } else { // Found is NOT final element
+            StrCpy(Found, &Comma[1]);
+        } // if/else
+        return TRUE;
+    } else {
+        return FALSE;
+    } // if/else
+} // BOOLEAN DeleteItemFromCsvList()
+
 // Returns TRUE if SmallString is an element in the comma-delimited List,
 // FALSE otherwise. Performs comparison case-insensitively.
 BOOLEAN IsIn(IN CHAR16 *SmallString, IN CHAR16 *List) {
@@ -335,6 +388,34 @@ BOOLEAN IsInSubstring(IN CHAR16 *BigString, IN CHAR16 *List) {
    return Found;
 } // BOOLEAN IsSubstringIn()
 
+// Replace *SearchString in **MainString with *ReplString -- but if *SearchString
+// is preceded by "%", instead remove that character.
+// Returns TRUE if replacement was done, FALSE otherwise.
+BOOLEAN ReplaceSubstring(IN OUT CHAR16 **MainString, IN CHAR16 *SearchString, IN CHAR16 *ReplString) {
+    BOOLEAN WasReplaced = FALSE;
+    CHAR16 *FoundSearchString, *NewString, *EndString;
+
+    FoundSearchString = MyStrStr(*MainString, SearchString);
+    if (FoundSearchString) {
+        NewString = AllocateZeroPool(sizeof(CHAR16) * StrLen(*MainString));
+        if (NewString) {
+            EndString = &(FoundSearchString[StrLen(SearchString)]);
+            FoundSearchString[0] = L'\0';
+            if ((FoundSearchString > *MainString) && (FoundSearchString[-1] == L'%')) {
+                FoundSearchString[-1] = L'\0';
+                ReplString = SearchString;
+            } // if
+            StrCpy(NewString, *MainString);
+            MergeStrings(&NewString, ReplString, L'\0');
+            MergeStrings(&NewString, EndString, L'\0');
+            MyFreePool(MainString);
+            *MainString = NewString;
+            WasReplaced = TRUE;
+        } // if
+    } // if
+    return WasReplaced;
+} // BOOLEAN ReplaceSubstring()
+
 // Returns TRUE if *Input contains nothing but valid hexadecimal characters,
 // FALSE otherwise. Note that a leading "0x" is NOT acceptable in the input!
 BOOLEAN IsValidHex(CHAR16 *Input) {
@@ -456,3 +537,15 @@ EFI_GUID StringAsGuid(CHAR16 * InString) {
 
     return Guid;
 } // EFI_GUID StringAsGuid()
+
+// Delete the STRING_LIST pointed to by *StringList.
+VOID DeleteStringList(STRING_LIST *StringList) {
+    STRING_LIST *Current = StringList, *Previous;
+
+    while (Current != NULL) {
+        MyFreePool(Current->Value);
+        Previous = Current;
+        Current = Current->Next;
+        MyFreePool(Previous);
+    }
+} // VOID DeleteStringList()
